/*
 * libxr16.c -- minimal xr16 runtime library
 *
 * Copyright (C) 1999, 2000, Gray Research LLC.  All rights reserved.
 * The contents of this file are subject to the XSOC License Agreement;
 * you may not use this file except in compliance with this Agreement.
 * See the LICENSE file.
 */

/*
 * XR16 library extensions - 
 *
 * Bit-shifts : sllr, srlr, srar 
 * Multiplication : muli2
 * Divide : divi2, divu2
 * Modulus: modu2, modi2
 *
 * This file must be compiled with a 'special' version of lcc-xr16 and
 * xr16asm.  The original lcc-xr16 compiler does not preserve
 * temporary registers for function-calls, nor does it correctly implement
 * variable bitshifts.
 *
 * liaor@iname.com : 06/08/2001
 */
  
typedef unsigned Word;

extern int main();
extern Word	_end;

void _zeromem();
Word* _tos();

/* Processor reset: zero memory and call main.
 */
int _reset() {
	_zeromem();

	main();

	/* dynamic halt */
	for (;;)
		;
}

/* Reset memory to consistent known state.
 */
void _zeromem() {
	Word* p = &_end;
	Word* end = _tos();

	for ( ; p < end; ++p)
		*p = 0;
}

/* Interrupts have not been tested.
 */
int _interrupt() {
	return 0;
}

/* Too clever way to return address of approximately the top of stack;
 * ignore compiler warning on returning address of argument.
 */
Word* _tos(Word arg) {
	return &arg;
}

/* Multiply unsigned times unsigned.
 */
unsigned mulu2(unsigned a, unsigned b) {
	unsigned w = 0;

	for ( ; a; a >>= 1) {
		if (a&1) w += b;
		b <<= 1;
	}
	return w;
}

/* Multiply signed times signed
 */
short signed muli2(short signed a, short signed b) {
	signed short ctr; /* countdown to MostSignificant Bit */
	signed short w = 0;

	for ( ctr=15; ((a) && (ctr>0)) ; --ctr ) {
		if (a&1)
			w += b;
		b <<= 1;
		a >>= 1;
	}

	/* This final check is on the SIGN (MSB) bit.*/
	if (a&1)
		w -= b;

	return w;
}

/* Shift left logical (register amount)
 */
unsigned sllr(unsigned value, unsigned amount) {
	
	for ( ; (amount > 0) ; --amount) {
		value = value << 1;
	}

	return value;
}

/* Shift right logical (register amount)
 */
unsigned srlr(unsigned value, unsigned amount) {
	
	for ( ; (amount > 0) ; --amount) {
		value = value >> 1;
	}

	return value;
}


/* Shift right arithmetic (register amount)
 */
signed srar(signed value, unsigned amount) {
	
	for ( ; (amount > 0) ; --amount) {
		value = value >> 1;
	}

	return value;
}


/* mask for divide operation - the mask selects a
   single bit from the operand */
const unsigned int mask_table[] = 
{
	0x0000,
	0x0001,
	0x0002,
	0x0004,
	0x0008,
	0x0010,
	0x0020,
	0x0040,
	0x0080,
	0x0100,
	0x0200,
	0x0400,
	0x0800,
	0x1000,
	0x2000,
	0x4000,
	0x8000
};

/* returns "X / Y" */
unsigned short int 
divu2( const unsigned short int x, const unsigned short int y )
{
	unsigned short int remainder;
	unsigned short int x_rmask;
	unsigned short int z;
	signed short int count;
	unsigned short int x_width = 16;
	unsigned short int y_width = 16;

	if ( y == 0 )
		return 0; /* DIVIDE_BY_0 : automatic failure */

	/* 1) determine the 'bit-widths' of X and Y */
	/*    Bit width is determined by the leading '1' bit */
	z = y;
	if ( z == 0 )
		y_width = 0;
	else
		while ( (z & 0x8000U)==0 )
		{
			z = z << 1;
			--y_width;
		}

	z = x;
	if ( z == 0 )
		x_width = 0;
	else
		while ( (z & 0x8000U)==0 )
		{
			z = z << 1;
			--x_width;
		}

	/* early exit?  Only if we know in advance that the result is 0! */
	if ( y_width > x_width )
		return 0;

	/* 2) initialize the division-engine */
	z = 0;
	if ( x_width >= y_width )
	{
		/* count = # iterations to to turn the division-engine */
		/* x_rmask= initial 'mask value', for shifting in the X bits */
		/* initial remainder value = 'right-justified' X argument */
		count = x_width - y_width;
		x_rmask = mask_table[ count ];
		remainder = x >> ( x_width-y_width);
	}
	else
	{
		/* because of 'exit early', this code should never execute!  */
		count = 0;
		x_rmask = mask_table[ 0 ];
		remainder = x;
	}


	/* 3) the division-engine! */
	while ( count >= 0 )
	{
		z = (z << 1);

		if ( y <= remainder )
		{
			remainder = remainder - y;

			/* write output '1' to Z-output register */
			z = z | (0x0001U);
		}

		if ( count > 0 )
		{
			remainder = (remainder << 1);

			/* pull in next-bit */
			if ( (x & x_rmask) != 0  )
				remainder = remainder | (0x0001U);

			x_rmask = (x_rmask >> 1);
		}

		--count;
	}

	return z;
}


/* returns "X % Y" */
unsigned short int 
modu2( const unsigned short int x, const unsigned short int y )
{
	unsigned short int remainder;
	unsigned short int x_rmask;
	unsigned short int z;
	signed short int count;
	unsigned short int x_width = 16;
	unsigned short int y_width = 16;

	if ( y == 0 )
		return 0; /* DIVIDE_BY_0 : automatic failure */

	/* 1) determine the 'bit-widths' of X and Y */
	/*    Bit width is determined by the leading '1' bit */
	z = y;
	if ( z == 0 )
		y_width = 0;
	else
		while ( (z & 0x8000U)==0 )
		{
			z = z << 1;
			--y_width;
		}

	z = x;
	if ( z == 0 )
		x_width = 0;
	else
		while ( (z & 0x8000U)==0 )
		{
			z = z << 1;
			--x_width;
		}

	/* early exit?  Only if we know in advance that the result == x! */
	if ( y_width > x_width )
		return x;

	/* 2) initialize the division-engine */
	if ( x_width >= y_width )
	{
		/* count = # iterations to to turn the division-engine */
		/* x_rmask= initial 'mask value', for shifting in the X bits */
		/* initial remainder value = 'right-justified' X argument */
		count = x_width - y_width;
		x_rmask = mask_table[ count ];
		remainder = x >> ( x_width-y_width);
	}
	else
	{ 
		/* because of 'exit early', this code should never execute!  */
		count = 0;
		x_rmask = mask_table[ 0 ];
		remainder = x;
	}


	/* 3) the division-engine! */
	while ( count >= 0 )
	{

		if ( y <= remainder )
		{
			remainder = remainder - y;

		}

		if ( count > 0 )
		{
			remainder = (remainder << 1);

			/* pull in next-bit */
			if ( (x & x_rmask) != 0  )
				remainder = remainder | (0x0001U);

			x_rmask = (x_rmask >> 1);
		}

		--count;
	}

	return remainder;
}


signed short int 
divi2( const signed short int x, const signed short int y )
{
	unsigned int u_x, u_y, u_z;
	unsigned sign_flag;

	if ( x < 0 )
	{
		sign_flag = 1;
		u_x = -x;
	}
	else
	{
		sign_flag = 0;
		u_x = x;
	}

	if ( y < 0 )
	{
		sign_flag = sign_flag ^ 1;
		u_y = -y;
	}
	else
		u_y = y;

	u_z = divide2( u_x, u_y );

	if ( sign_flag )
		return (-u_z);
	else
		return u_z;
}


signed short int 
modi2( const signed short int x, const signed short int y )
{
	unsigned int u_x, u_y, u_z;
	unsigned sign_flag;

	if ( x < 0 )
	{
		sign_flag = 1;
		u_x = -x;
	}
	else
	{
		sign_flag = 0;
		u_x = x;
	}

	if ( y < 0 )
	{
		sign_flag = sign_flag;
		u_y = -y;
	}
	else
		u_y = y;

	u_z = modulus2( u_x, u_y );

	if ( sign_flag )
		return (-u_z);
	else
		return u_z;
}

/* This must be the last symbol in the last module "linked" into the load
 * image.
 */
Word _end;
